## heroku-buildpack-sigsci

This is an initial stab at packaging the agent and a golang reverse proxy into
an Heroku buildpack.

Consider this experimental and a bit of a proof of concept. These docs are focussed
initially on how to get started to hack on this.

## Buildpacks how do they work?

Heroku build packs aren't anything more special than a tarball and a few shell
scripts. When you plumb a "buildpack" into your heroku app the following happens on deploy:

* your dyno gets wiped and reprovisioned
* heroku pulls the build pack tarball down
* the tarball gets unpacked into the dyno
* the `bin/detect` script gets run - most buildpacks just echo a line saying hello, this
script can be used to detect whether the buildpack actually needs to be compiled/run
* the `bin/compile` script is run - this does all the work, essentially this is the place
you put all your code to bootstrap and configure your buildpack
* the `bin/release` script gets called - this just outputs metadata back to the runtime in 
the format of a YAML file

For the most part compile is all we care about for this one.  The buildpack docs can be
found at [https://devcenter.heroku.com/articles/buildpack-api](https://devcenter.heroku.com/articles/buildpack-api)

### Multiple Buildpacks

By default a heroku app only supports a single buildpack and it will detect from your
application source which buildpack to use. These buildpacks are typically the build tools
for your given application. For instance, if you push up an app that looks like golang code
it will use the heroku golang buildpack.

For less trivial apps folks may want to run more than one buildpack. In this case the 
recommended approach is to use the multi-buildpack buildpack. This allows you to set `buildpack-multi` as your buildpack then in the top level of your application source you can list all
the buildpacks you want to use, it will then pull each of them in-order and run their compile
step.

Docs for buildpack-multi are here [https://github.com/ddollar/heroku-buildpack-multi](https://github.com/ddollar/heroku-buildpack-multi)

## Sigsci Buildpack

### Heroku architecture 30 sec primer

The heroku environment imposes some interesting operational characteristics. A typical web
application consists of one or more dynos. A dyno is essentially a resource controlled 
container for running your web application code.

Dynos typically run a single threaded web application and scaling inbound request capacity
is performed by spinning up more dynos. The heroku API makes this pretty trivial to do.

The web application dynos don't face the internet directly, when a new app is created
Heroku setup an internet facing load balancer for inbound traffic that then routes requests
to your application dynos. This is all magic and opaque - it just works.

The application load balancer terminates TLS on the edge and all origin traffic to the dynos
is plain HTTP, yay!

The heroku architecture follows the [twelve factor app](http://12factor.net/) model. In
parcitular apps are stateless, configuration is via environment and inter-process communication
(across dyno boundaries) is via an offically provided intermediary persistence layer (postgres / redis / rabbit mq). Arbitrary TCP/IP connections between dynos are not possible.

### Sigsci on a dyno

Some assumptions have been made on what a typical Heroku app deployment looks like. Given 
we can't communicate over TCP between dynos this initial model has us run multiple
processes on a single web dyno.

The goal is to require little to no modification to the customer's application and allow us
to get plumbed in with nothing more than adding the buildpack and setting a few configuration
values.

To achieve this we drop a small golang reverse proxy into the dyno this proxies inbound requests
to the customers application. The request flow looks like:

1. client makes request
1. request hits heroku load balancer
1. load balancer forwards request to dyno
1. proxy accepts request
1. request is sent to the agent over the UNIX domain socket (golang module)
1. module acts on agent response
1. request is forwarded to the customer's application

The main change required to the customer's application is for them to:

* modify their application to bind to an ephemeral port on localhost (eg. 127.0.0.1:1234)
* set the environment variable SIGSCI_UPSTREAM to that address (eg. `SIGSCI_UPSTREAM=127.0.0.1:1234`)
* modify their application Procfile to call `bin/sigsci-start <their app command line>`

  For example, if a customer's Procfile looked like this:

    ```
    web: bin/my-awesome-web-app
    ```

  After changing it would look like:

    ```
    web: sigsci/bin/sigsci-start bin/my-awesome-web-app
    ```

The `sigsci-start` script is a little shell script wrapper that:

1. starts the agent
1. starts the reverse proxy
1. starts the customer's application using the command passed on the command line

We have more options available here. If folks run the language specific module implementations we could ditch the reverse proxy.

#### Not a free lunch ...

Concerns right now are the performance and stability of running two additional processes
inside a single dyno. This needs further exploration. Running the agent alongside the 
web app is likely to increase throughput and CPU utilization.

There's also an open issue on the golang ReverseProxy implementation, there's concerns
it is not performant, and from some intiial benchmarking it appears to introduce additional
latency. See: https://github.com/golang/go/issues/10277

More work needed here.

## Wiring us into an Heroku app

1. At the top of the repo for the app you want to deploy create a new heroku app:

    ```
    heroku create
    ```

   This calls the heroku API to create a new app and adds a new git remote origin
   to the git configuration for the repo.

1. Set the buildpack for the app to buildpack-multi more detail in the section below:

    ```
    heroku buildpacks:set https://github.com/ddollar/heroku-buildpack-multi.git
    ```

1. Create a .buildpacks file listing the buildpacks you want to use. In this example
   we add the sigsci buildpack and we also add the golang buildpack (our app is a golang app).
   For other platforms you'd use the appropriate language build pack (ruby, php etc...)

    ```
    https://s3-us-west-2.amazonaws.com/signalsciences-artifacts-us-west-2/sigsci/heroku-buildpack-sigsci.tgz
    https://github.com/heroku/heroku-buildpack-go.git
    ```

1. Set the following application configuration variables, these are stored alongside the
   app and are made available to the buildpacks in the `ENV_DIR` at build time:

    ```
    heroku config:set SIGSCI_ACCESS_KEY=<your access key>
    heroku config:set SIGSCI_SECRET_KEY=<your secret>
    ```

1. Configure your application to bind to localhost on an ephemeral port and configure
   this address as the upstream for the sigsci proxy:

    ```
    heroku config:set SIGSCI_UPSTREAM=<your app listener>
    ```

   eg:
    ```
    heroku config:set SIGSCI_UPSTREAM=127.0.0.1:8888
    ```

1. Deploy your app:

    ```
    git push heroku master
    ```

   This will push your code up to heroku and trigger a deploy, this will download and
   configure the agent and reverse proxy, once the deploy is complete your app should be 
   up and listening, all inbound requests will be routed via the proxy and via the agent.

## Hacking on the buildpack

To hack on the buildpack clone this repo, make your changes then run `./release_s3.sh`. The
release script tars up the buildpack and pushes to the s3 artifact bucket. There's no versioning
so a release just overwrites the last version.

Once we decide we're close to something shippable we'll build in a proper release process.

You'll need awscli installed to release.
